# Copyright (c) 2016 Rob Ruana
# Licensed under the MIT License, see LICENSE for details.

from __future__ import absolute_import
import os
import six

from collections import OrderedDict
from difflib import SequenceMatcher
from lxml import html
from math import ceil, floor
from PIL import Image, ImageChops, ImageEnhance

from hudsucker.industries import (
    chunks,
    circular_iter,
    filesystem_safe,
    is_imageish,
    is_urlish,
    log,
    scale_to_fit,
    split_url_filename,
    url_filename,
    DownloadCache)


class CardInfo(object):

    @classmethod
    def parse(cls, file_path):
        with open(file_path) as file:
            lines = [line.partition('#')[0].strip() for line in file]
            return [cls(line) for line in lines if line]

    def __init__(self, card):
        self.raw = card
        if is_urlish(card):
            self.name, _ = split_url_filename(card)
            if is_imageish(card):
                self.html_url = None
                self.image_url = card
            else:
                self.html_url = card
                self.image_url = None
        else:
            self.name = card
            self.html_url = None
            self.image_url = None
        self.image_basename = filesystem_safe(self.raw)

    def guess_filename(self):
        if self.image_url:
            return url_filename(self.image_url)
        elif self.html_url:
            return url_filename(self.html_url)
        else:
            return filesystem_safe(self.name)


class HulaHoop(object):

    search_engines = OrderedDict([
        ('ask', ('http://www.ask.com/web?q=', 'a')),
        ('bing', ('http://www.bing.com/search?q=', 'a')),
        ('duckduckgo', ('https://duckduckgo.com/html/?q=', 'a')),
        ('yahoo', ('https://search.yahoo.com/search?p=', 'a'))])

    search_engines_round_robin = circular_iter(list(search_engines.keys()))

    def __init__(self, border, quality, resolution, site, output_dir, cache_dir, no_cache, refresh_cache):
        self.quality = quality
        self.resolution = resolution
        self.site = site
        self.output_dir = output_dir
        self.cache_dir = os.path.join(cache_dir, '_download_cache_')
        self.no_cache = no_cache
        self.refresh_cache = refresh_cache
        self.cache = None  # Cache is initialized in run() method

        # Empirically determined inner card size: 2.24 inches, 3.24 inches
        border = round(border * resolution * 2.0) / 2.0
        self.border_leading = int(floor(border))
        self.border_trailing = int(ceil(border))
        self.inner_card_width = int(round(2.4 * resolution))
        self.inner_card_height = int(round(3.4 * resolution))
        self.outer_card_width = int(self.inner_card_width + self.border_leading + self.border_trailing)
        self.outer_card_height = int(self.inner_card_height + self.border_leading + self.border_trailing)

    def _crop_border(self, image):
        bright_image = ImageEnhance.Brightness(image).enhance(2)
        bg = Image.new(image.mode, image.size, image.getpixel((int(round(image.width / 2.0)), 0)))
        diff = ImageChops.difference(bright_image, bg)
        diff = ImageChops.add(diff, diff, 1, -100)
        bbox = diff.getbbox()
        return image.crop(bbox)

    def _image_file_for_card(self, card):
        cache_path = self.cache.get_cached_file(card.image_basename)
        if cache_path:
            log('Using cached image: {}'.format(cache_path))
            return cache_path

        log()
        indention = log.indention
        if not card.image_url:
            if not card.html_url:
                log('Searching for "{}" on {}...'.format(card.name, self.site))
                card.html_url = self._search_for_card(card)
                log.indention += 4

            log('Looking for image in: {}'.format(card.html_url))
            card.image_url = self._image_url_from_html(card)
            log.indention += 4

        log('Downloading: {}'.format(card.image_url))
        with log.indent(blank_line=True):
            result = self.cache.get_file(card.image_url, cache_basename=card.image_basename)
        log.indention = indention
        return result

    def _images_for_cards(self, cards):
        images = []
        for card in cards:
            image_path = self._image_file_for_card(card)
            if image_path:
                images.append((Image.open(image_path), card))
            else:
                six.print_('')
                six.print_('Could not find image for: "{}"'.format(card.raw))
                six.print_('')
        return images

    @log.auto_indent()
    def _image_url_from_html(self, card):
        text = self.cache.get_text(card.html_url)
        if not text:
            return None
        doc = html.fromstring(text)
        urls = [m.get('content').strip() for m in doc.cssselect('meta') if m.get('property') == 'og:image']
        image_url = urls[0] if urls else None
        image_path_url, _, _ = card.html_url.rpartition('/')
        if image_url:
            if image_url.startswith('http'):
                log('Found image url in meta tag: {}'.format(image_url))
            else:
                log('Found relative image url in meta tag: {}'.format(image_url))
                image_url = '{}/{}'.format(image_path_url, image_url)
        else:
            urls = OrderedDict()
            for i in doc.cssselect('img'):
                src = i.get('src', '').strip()
                if src:
                    urls[src] = i.get('alt', '').strip()

            log('Didn\'t find image url in meta tag, searching html...')
            image_url = self._url_matching_card(card, urls)
            if not image_url:
                image_url = '{}/{}.jpg'.format(image_path_url, card.guess_filename())
                log('Didn\'t find image url in html, guessing: {}'.format(image_url))
        return image_url

    def _process_card_file(self, input_path):
        cards = CardInfo.parse(input_path)

        sheet_basename, _ = os.path.splitext(os.path.basename(input_path))
        sheets = []
        for (sheet_index, sheet) in enumerate(chunks(self._images_for_cards(cards), 9)):
            cropped_images = []
            for (image, card) in sheet:
                cropped_image = self._crop_border(image)
                cropped_images.append((cropped_image, card))

            card_count = len(sheet)
            sheet_width = self.outer_card_width * min(3.0, card_count)
            sheet_height = self.outer_card_height * ceil(card_count / 3.0)

            sheet_filename = '{}{:02d}.pdf'.format(sheet_basename, sheet_index + 1)
            sheet_path = os.path.join(self.output_dir, sheet_filename)
            sheet_image = Image.new('RGB', (int(sheet_width), int(sheet_height)), 'white')

            for (card_index, (image, card)) in enumerate(cropped_images):
                if image.width != self.inner_card_width or image.height != self.inner_card_height:
                    new_width, new_height = scale_to_fit(image.size, (self.inner_card_width, self.inner_card_height))
                    new_width, new_height = int(ceil(new_width)), int(ceil(new_height))
                    if new_width != image.width and new_height != image.height:
                        image = image.resize((new_width, new_height), Image.LANCZOS)

                inner_card_x = max(self.border_leading, floor((self.outer_card_width - image.width) / 2.0))
                inner_card_y = max(self.border_leading, floor((self.outer_card_height - image.height) / 2.0))
                border_image = Image.new('RGB', (self.outer_card_width, self.outer_card_height), 'black')
                border_image.paste(image, (int(inner_card_x), int(inner_card_y)))

                outer_card_x = self.outer_card_width * (card_index % 3.0)
                outer_card_y = self.outer_card_height * floor(card_index / 3.0)
                sheet_image.paste(border_image, (int(outer_card_x), int(outer_card_y)))

            log('\nSaving: {}\n'.format(sheet_path))
            sheets.append(sheet_path)
            sheet_image.save(sheet_path, 'PDF', quality=self.quality,
                             resolution=self.resolution, resolution_unit='inch')
        return sheets

    @log.auto_indent()
    def _search_for_card(self, card):
        for engine in self.search_engines_round_robin:
            url, css_selector = self.search_engines[engine]
            text = self.cache.get_text(url, '{} site:{}'.format(card.name, self.site))
            if text:
                doc = html.fromstring(text)
                urls = OrderedDict()
                for a in doc.cssselect(css_selector):
                    href = a.get('href', '').strip()
                    if href and href.startswith('http') and self.site in href and href not in urls:
                        urls[href] = a.text_content().strip()

                result = self._url_matching_card(card, urls)
                if result:
                    return result
        return None

    def _url_matching_card(self, card, urls, threshold=0.7):
        results = []
        source_filename = split_url_filename(card.guess_filename())[0].lower()
        image_filename = card.image_basename.lower()
        card_name = card.name.lower()
        for (url, text) in urls.items():
            text = text.lower()
            url_filename = split_url_filename(url)[0].lower()
            image_ratio = SequenceMatcher(None, image_filename, url_filename).ratio()
            source_ratio = SequenceMatcher(None, source_filename, url_filename).ratio()
            if card_name in text:
                card_ratio = 0.9
            else:
                card_ratio = SequenceMatcher(None, card_name, text).ratio()
            ratio = max(image_ratio, card_ratio, source_ratio)
            if ratio >= threshold:
                results.append((url, ratio))

        log('Found {} result{}:'.format(len(results), '' if len(results) == 1 else 's'))
        matching_result = None
        matching_ratio = 0
        for (result, ratio) in results:
            with log.indent(): log('{} (Matches {:.2f}%)'.format(result, ratio * 100))
            if ratio > matching_ratio:
                matching_result = result
                matching_ratio = ratio
        if len(results) > 1:
            log('Using closest matching result: {}'.format(matching_result))

        return matching_result

    def run(self, input_paths):
        if not os.path.exists(self.output_dir):
            log('Creating output dir: {}\n'.format(self.output_dir))
            os.makedirs(self.output_dir)

        self.cache = DownloadCache(self.cache_dir, self.refresh_cache, self.no_cache, '_hudsucker_cache_')
        with self.cache.auto_cleanup():
            sheets = []
            for input_path in input_paths:
                sheets.extend(self._process_card_file(input_path))
            for sheet in sheets:
                six.print_(sheet)



#!/usr/bin/env python

# Copyright (c) 2016 Rob Ruana
# Licensed under the MIT License, see LICENSE for details.

from __future__ import absolute_import
import argparse
import os
import sys

from hudsucker.industries import log
from hudsucker.proxy import HulaHoop


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Hudsucker Proxy Generator - generate MtG proxy sheets')
    parser.add_argument('input', metavar='FILE', nargs='+',
                        help='each line of FILE should be a MtG card name, or a url')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        help='print verbose details')

    output_group = parser.add_argument_group('proxy sheet options')
    output_group.add_argument('-b', '--border', dest='border', metavar='N', default=0.072, type=float,
                              help='border width in inches, defaults to 0.072')
    output_group.add_argument('-o', '--output', dest='output_dir', metavar='DIR', default='.',
                              help='output dir, defaults to current dir')
    output_group.add_argument('-p', '--resolution', dest='resolution', metavar='N', default=600.0, type=float,
                              help='print resolution of output PDF, defaults to 600')
    output_group.add_argument('-q', '--quality', dest='quality', metavar='N', default=95, type=int,
                              help='quality to use for JPEG encoding, defaults to 95')
    output_group.add_argument('-s', '--site', dest='site', metavar='URL', default='mythicspoiler.com',
                              help='site to search for card images, defaults to mythicspoiler.com')

    cache_group = parser.add_argument_group('caching options', description='NOTE: '
                                            'Careful turning off cache, search engines may ban your IP')
    cache_group.add_argument('-c', '--cache', dest='cache_dir', metavar='DIR', default='hudsucker_cache',
                             help='cache dir, defaults to hudsucker_cache')
    cache_group.add_argument('-n', '--no-cache', dest='no_cache', action='store_true',
                             help='don\'t cache any downloaded files')
    cache_group.add_argument('-r', '--refresh', dest='refresh_cache', action='store_true',
                             help='force refresh of any cached downloads')

    args = parser.parse_args(argv[1:])

    log.verbose = args.verbose

    hula_hoop = HulaHoop(
        border=args.border,
        quality=min(95, max(1, args.quality)),
        resolution=args.resolution,
        site=args.site,
        output_dir=os.path.expanduser(args.output_dir),
        cache_dir=os.path.expanduser(args.cache_dir),
        no_cache=args.no_cache,
        refresh_cache=args.refresh_cache)

    hula_hoop.run(args.input)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2016 Rob Ruana
# Licensed under the MIT License, see LICENSE for details.

from __future__ import absolute_import
import glob
import os
import re
import requests
import shutil
import six
import tempfile
import time

from contextlib import contextmanager
from functools import wraps
from six.moves.urllib.parse import quote, urlparse


def collapse_blank_lines(text, pre_count=1, post_count=1):
    pre_lines = []
    first_line, _, text_body = text.partition('\n')
    while not first_line.strip() and text_body:
        text = text_body
        if len(pre_lines) < pre_count:
            pre_lines.append(first_line)
        first_line, _, text_body = text.partition('\n')

    post_lines = []
    text_body, _, last_line = text.rpartition('\n')
    while not last_line.strip() and text_body:
        text = text_body
        if len(post_lines) < post_count:
            post_lines.append(last_line)
        text_body, _, last_line = text.rpartition('\n')

    return '\n'.join(pre_lines + [text] + post_lines)


def indent(text, prefix):
    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if line.strip() else line)
    return ''.join(prefixed_lines())


class VerbosityLogger(object):

    def __init__(self, verbose=False, collapse_blank_lines=True):
        self.verbose = verbose
        self.collapse_blank_lines = collapse_blank_lines
        self.has_blank_line = True
        self.indention = 0

    def __call__(self, message=''):
        if self.verbose:
            if self.collapse_blank_lines:
                pre_count = 0 if self.has_blank_line else 1
                message = collapse_blank_lines(message, pre_count)
                if self.has_blank_line and not message.strip():
                    return
            if self.indention > 0:
                six.print_(indent(message, ' ' * self.indention))
            else:
                six.print_(message)
            self.has_blank_line = message.rsplit('\n', 1)[-1].strip() == ''

    def auto_indent(self, n=4, blank_line=False):
        def outer(func):
            @wraps(func)
            def inner(*a, **kw):
                with self.indent(n, blank_line):
                    return func(*a, **kw)
            return inner
        return outer

    def auto_log(self, message=''):
        def outer(func):
            @wraps(func)
            def inner(*a, **kw):
                result = func(*a, **kw)
                self(message)
                return result
            return inner
        return outer

    @contextmanager
    def indent(self, n=4, blank_line=False):
        self.indention += n
        yield self
        self.indention -= n
        if blank_line:
            self()


log = VerbosityLogger()


def chunks(iterable, size=2):
    return [iterable[x:x + size] for x in range(0, len(iterable), size)]


class circular_iter(object):

    def __init__(self, indexable):
        self.nextIndex = 0
        self.stopIndex = -1
        self.indexable = indexable

    def __iter__(self):
        self.stopIndex = -1
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self.stopIndex == -1:
            self.stopIndex = self.nextIndex
        elif self.stopIndex == self.nextIndex:
            raise StopIteration()
        nextItem = self.indexable[self.nextIndex]
        self.nextIndex = (self.nextIndex + 1) % len(self.indexable)
        return nextItem


def filesystem_safe(s):
    return re.sub(r'[^\w\.\-_]+', '_', s.encode('ascii', 'ignore').decode('utf-8'))


def is_imageish(s):
    s = s.lower()
    return (s.endswith('.bmp') or s.endswith('.gif') or s.endswith('.jpg') or
            s.endswith('.jpeg') or s.endswith('.png') or s.endswith('.tga') or
            s.endswith('.tif') or s.endswith('.tiff') or s.endswith('.webp'))


def is_urlish(s):
    s = s.lower()
    return s.startswith('http') or (s.count('.') >= 2 and s.count('/') >= 1)


def scale_to_fit(size, bounds):
    width, height = size
    bounds_width, bounds_height = bounds
    if width <= 0:
        return (width, min(height, bounds_height))
    if height <= 0:
        return (min(width, bounds_width), height)
    # get true-division on python 2
    scale = min(float(bounds_width) / width, float(bounds_height) / height)
    return (width * scale, height * scale)


def split_url_filename(url):
    if url:
        return os.path.splitext(url_filename(url))
    return '', ''


def url_filename(url):
    if url:
        return os.path.basename(urlparse(url).path)
    return ''


class DownloadCache(object):

    def __init__(self, cache_dir, should_refresh, is_temp, temp_prefix='_download_cache_'):
        self.should_refresh = should_refresh
        self.is_temp = is_temp
        if is_temp:
            self.temp_dir = tempfile.mkdtemp(prefix=temp_prefix)
            self.cache_dir = self.temp_dir
            log('Using temp cache dir: {}\n'.format(self.cache_dir))
        else:
            self.temp_dir = None
            self.cache_dir = cache_dir
            if self.should_refresh:
                if os.path.exists(self.cache_dir):
                    log('Refreshing cache dir: {}\n'.format(self.cache_dir))
                    shutil.rmtree(self.cache_dir)
                    os.makedirs(self.cache_dir)
            if not os.path.exists(self.cache_dir):
                log('Creating cache dir: {}\n'.format(self.cache_dir))
                os.makedirs(self.cache_dir)

    @contextmanager
    def auto_cleanup(self):
        yield self
        self.cleanup()

    def cleanup(self):
        if self.temp_dir and os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)

    def get_cached_file(self, cache_basename):
        for file in glob.glob(os.path.join(self.cache_dir, cache_basename + '.*')):
            return file
        return None

    def get_file(self, url, query='', cache_basename=None):
        if not url:
            return None

        log('GET \'{}{}\''.format(url, query))
        query_url = url + quote(query)

        if cache_basename:
            _, ext = split_url_filename(url)
            cache_filename = '{}{}'.format(cache_basename, ext)
        else:
            cache_filename = filesystem_safe(query_url)
        cache_path = os.path.join(self.cache_dir, cache_filename)

        if os.path.exists(cache_path):
            with log.indent(): log('Using cached file: {}'.format(cache_path))
            return cache_path
        else:
            response = requests.get(query_url, stream=True)
            if response.ok:
                with open(cache_path, 'wb') as cache_file:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            cache_file.write(chunk)
            else:
                cache_path = None
                six.print_('')
                six.print_('ERROR {} {}'.format(response.status_code, response.reason))
                six.print_(response.text)
                six.print_('')

            # Take it easy on those servers out there :)
            time.sleep(0.5)
            return cache_path

    def get_text(self, url, query='', cache_basename=None):
        cache_path = self.get_file(url, query, cache_basename)
        text = None
        if cache_path and os.path.exists(cache_path):
            with open(cache_path, 'r') as cache_file:
                text = cache_file.read()
        return text
